// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: CachedIndexFile.cc,v 0.9 1997-12-27 07:07:12+09 harada Exp $
//
extern "C" {
#include <assert.h>
}
#include "config.h"
#include "CachedIndexFile.h"

CachedIndexFile::CachedIndexFile(int maxpage, fstream &file,
                                 int rdonlyp, int newfilep)
    : IndexFile(file, rdonlyp, newfilep),
      maxpageNum(maxpage), pageNum(0), sync_need(0)
{
    lruQ = new int[maxpageNum];
    entries = new (struct cache_entry_t)[maxpageNum];
    for(int i = 0; i < maxpageNum; i++){
        lruQ[i] = -1;
        entries[i].buffer = NULL;
        entries[i].bn = 0;
        entries[i].size = 0;
    }
}

CachedIndexFile::~CachedIndexFile(){
    if(sync_need) sync();
    for(int i = 0; i < pageNum; i++){
        assert(entries[i].buffer != NULL);
        delete [] entries[i].buffer;
    }
    delete []entries;
    delete []lruQ;
}

void
CachedIndexFile::sync(){
    for(int p = 0; p < pageNum; p++)
        IndexFile::writeBlock(entries[p].buffer,
                              entries[p].bn, entries[p].size);
}

int
CachedIndexFile::findPage(blockno_t bn){
    for(int i = 0; i < pageNum; i++){
        assert(NULL != entries[i].buffer);
        if(entries[i].bn == bn) return i;
    }
    return -1;                  // not found
}

int
CachedIndexFile::liftup(int p, int hint){
    int inq;                    // lruQinqܤ־˻äƤ
    if(hint == -1){
        inq = -1;
        for(int i = 0; i < pageNum; i++){
            if(lruQ[i] == p){ inq = i; break; }
        }
    }else{
        assert(lruQ[hint] == p);
        inq = hint;
    }
    assert(inq != -1);
    if(inq >= 1){
        for(int j = inq; j > 0; j--)
            lruQ[j] = lruQ[j - 1];
        lruQ[0] = p;
    }
    return inq;
}

int
CachedIndexFile::pushout(int &hint){
    assert(pageNum == maxpageNum);
    hint = maxpageNum - 1;
    int p = lruQ[hint];
    if(sync_need)
        IndexFile::writeBlock(entries[p].buffer,
                              entries[p].bn, entries[p].size);
    return p;
}

int
CachedIndexFile::readBlock(char *buf, blockno_t bn, size_t size){
#if 0
    return IndexFile::readBlock(buf, bn, size);
#else
    int slot = findPage(bn);
    if(-1 != slot){             // cache hit
        assert(entries[slot].size == size);
        liftup(slot);
        memcpy(buf, entries[slot].buffer, size);
    }else if(pageNum < maxpageNum){ // fault, we still have an entry in empty
        assert(entries[pageNum].buffer == NULL);
        entries[pageNum].buffer = new char[size];
        IndexFile::readBlock(entries[pageNum].buffer, bn, size);
        lruQ[pageNum] = pageNum;
        entries[pageNum].size = size;
        entries[pageNum].bn = bn;
        memcpy(buf, entries[pageNum].buffer, size);
        pageNum++;
    }else{                      // fault, we must make a space for new entry
        int hint = -1;
        slot = pushout(hint);
        assert(lruQ[hint] == slot);
        if(entries[slot].size < size){ // realloc memory
            delete [] entries[slot].buffer;
            entries[slot].buffer = new char[size];
        }
        IndexFile::readBlock(entries[slot].buffer, bn, size);
        memcpy(buf, entries[slot].buffer, size);
        entries[slot].size = size;
        entries[slot].bn = bn;
        liftup(slot, hint);
    }
    return ! idxfile.good();
#endif
}

int
CachedIndexFile::writeBlock(const char *buf, blockno_t bn, size_t size){
#if 0
    return IndexFile::writeBlock(buf, bn, size);
#else
    sync_need = 1;
    int p = findPage(bn);
    if(-1 == p){
        if(pageNum < maxpageNum){
            assert(entries[pageNum].buffer == NULL);
            entries[pageNum].buffer = new char[size];
            lruQ[pageNum] = pageNum;
            entries[pageNum].size = size;
            entries[pageNum].bn = bn;
            memcpy(entries[pageNum].buffer, buf, size);
            pageNum++;
        }else{
            int hint = -1;
            int slot = pushout(hint);
            assert(lruQ[hint] == slot);
            if(entries[slot].size < size){ // realloc memory
                delete [] entries[slot].buffer;
                entries[slot].buffer = new char[size];
            }
            memcpy(entries[slot].buffer, buf, size);
            entries[slot].size = size;
            entries[slot].bn = bn;
            liftup(slot, hint);
        }
    }else{
        assert(entries[p].size == size);
        memcpy(entries[p].buffer, buf, size);
    }
    return ! idxfile.good();;
#endif
}

