// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Description.cc,v 0.10 1998-01-08 01:45:36+09 harada Exp $
//
extern "C" {
#include <assert.h>
#include <stdlib.h>
#include <time.h>
#include <sys/param.h>          // ntohl(3)
#include <netinet/in.h>         // ntohl(3) (Linux, etc)
}
#include "config.h"
#include "Description.h"
#include "htmlutil.h"

// TODO: numerical timezone is prefererd (RFC 1123)
const char *
Description::rfc822date(char *buf, size_t bufsiz, time_t &t){
    (void)strftime(buf, bufsiz, "%a, %d %b %Y %H:%M:%S %Z", localtime(&t));
    return buf;
}

void
Description::getblock(int n, time_t &date, time_t &lastmodified,
                      size_t &sz, char *url,
                      char *author, char *title, char *digest){
    url[0] = author[0] = title[0] = digest[0] = '\0';
    date = lastmodified = 0;
    sz = 0;

    char buf[blksize + 1];
    file.seekg(blksize * n);
    file.read(buf, blksize);
    if(file.gcount() != (size_t)blksize){
        PRINTERRMSG("read error.");
        return;
    }
    buf[blksize] = '\0';

    sz = ntohl(*(size_t *)buf);
    date = ntohl(*(time_t *)(buf + sizeof(sz)));
    lastmodified = ntohl(*(time_t *)(buf + sizeof(sz) + sizeof(date)));

    char *cp = buf + sizeof(sz) + sizeof(date) + sizeof(lastmodified);
    strcpy(url, cp);

    while(*cp && cp < buf + blksize - 1) cp++;
    strcpy(author, ++cp);

    while(*cp && cp < buf + blksize - 1) cp++;
    strcpy(title, ++cp);

    while(*cp && cp < buf + blksize - 1) cp++;
    strcpy(digest, ++cp);
}

void
Description::print(ostream &out, const Result &result){
    char path[blksize + 1];
    char author[blksize + 1];
    char title[blksize + 1];
    char digest[blksize + 1];
    char datebuf[128];
    time_t date, lastmod;
    size_t sz;

    getblock(result.docid, date, lastmod, sz, path, author, title, digest);
#ifndef NDEBUG
    out << "* docid = " << result.docid << endl;
#endif
    if(title[0])
        out << "Title:\"" << title << '"' << endl;
    if(path[0])
        out << "  File: " << path;
    out << " (score: " << (int)result.weight;
    out << ", size: " << sz << " bytes)" << endl;
    if(date != 0)
        out << "  Created: " << rfc822date(datebuf, 128, date) << endl;
    if(lastmod != 0)
        out << "  Modified: " << rfc822date(datebuf, 128, lastmod) << endl;
    if(author[0])
        out << "  Author: " << author << endl;
    if(digest[0])
        out << ">> " << digest << ".." << endl;
    else
        out << " (no summary)" << endl;
    out << endl; 
}

void
Description::printHtml(ostream &out, const Result &result){
    char url[blksize + 1];
    char author[blksize + 1];
    char title[blksize + 1];
    char digest[blksize + 1];
    time_t date, lastmod;
    size_t sz;

    getblock(result.docid, date , lastmod, sz, url, author, title, digest);
    char *escaped_url = escape_url(url);
    char buf[blksize + 1];
    char datebuf[MAXDATESTR + 1];

    out << "<!-- id:" << result.docid << "-->" << endl;
    out << "<DT><B><A HREF=\"" << escaped_url << "\">";
    if(title[0]){
        out << html_entitize(buf, title);
    }else out << "(Untitled)";
    out << "</A></B>\n";
    out << " <SMALL>(score: " << (int)result.weight << ")</SMALL>\n";
    out << "<DD>";
    if(author[0])
        out << " Author: <TT>"
            << html_entitize(buf, author) << "</TT><BR>\n";
    out << "&gt;&gt; " << html_entitize(buf, digest) << "..<BR>\n";
    out << "<SMALL><NOBR><A HREF=\"" << escaped_url << "\">"
        << escaped_url << "</A></NOBR> \n";
    out << "<NOBR>( " << sz << " bytes)</NOBR>";
    if(date != 0)
        out << " <NOBR>Created:"
            << rfc822date(datebuf, MAXDATESTR, date) << "</NOBR>";
    if(lastmod != 0)
        out << " <NOBR>Modified:"
            << rfc822date(datebuf, MAXDATESTR, lastmod) << "</NOBR>";
    out << "</SMALL><BR><BR>" << endl; 
    delete []escaped_url;
}


// ------------------------------------------------------------

void
HtmlDescription::print(ostream &out, const Result &result){
    char url[blksize + 1];
    char author[blksize + 1];
    char title[blksize + 1];
    char digest[blksize + 1];
    char datebuf[MAXDATESTR + 1];
    time_t date, lastmod;
    size_t sz;

    getblock(result.docid, date, lastmod, sz, url, author, title, digest);
#ifndef NDEBUG
    out << "* docid = " << result.docid << endl;
#endif
    out << '"' << title << '"';
    out << " (score: " << (int)result.weight;
    out << ", size: " << sz << " bytes)" << endl;
    out << "  URL: " << url << endl;
    out << "  Date: " << rfc822date(datebuf, MAXDATESTR, lastmod);
    if(author[0]){
        out << "  Author: " << author << endl;
    }
    out << ">> " << digest << ".." << endl;
    out << endl; 
}

void
HtmlDescription::printHtml(ostream &out, const Result &result){
    char url[blksize + 1];
    char author[blksize + 1];
    char title[blksize + 1];
    char digest[blksize + 1];
    time_t date, lastmod;
    size_t sz;

    getblock(result.docid, date, lastmod, sz, url, author, title, digest);
    char *escaped_url = escape_url(url);
    char buf[blksize + 1];
    char datebuf[MAXDATESTR + 1];

    out << "<!-- id:" << result.docid << "-->" << endl;
    out << "<DT><B><A HREF=\"" << escaped_url << "\">";
    if(title[0]){
        out << html_entitize(buf, title);
    }else out << "(Untitled)";
    out << "</A></B>\n";
    out << " <SMALL>(score: " << (int)result.weight << ")</SMALL>\n";
    out << "<DD>";
    if(author[0]){
        out << "  Author: <TT>"
            << html_entitize(buf, author) << "</TT><BR>\n";
    }
    out << "&gt;&gt; " << html_entitize(buf, digest) << "..<BR>\n";
    out << "<SMALL><NOBR><A HREF=\"" << escaped_url << "\">"
        << escaped_url << "</A></NOBR> \n";
    out << "<NOBR>( " << sz << " bytes)</NOBR> <NOBR>";
    out << rfc822date(datebuf, MAXDATESTR, lastmod)
        << "</NOBR></SMALL><BR><BR>\n";
    out << endl; 
    delete []escaped_url;
}

// ------------------------------------------------------------

void
MailDescription::print(ostream &out, const Result &result){
    char buf[blksize + 1];
    file.seekg(blksize * result.docid);
    file.read(buf, blksize);
    buf[blksize] = '\0';

    assert(sizeof(time_t) == sizeof(u_long));
    size_t sz = ntohl( *(size_t *)buf );
    time_t tm = ntohl( *(time_t *)(buf + sizeof(size_t)) );
    // time_t lm = ntohl( *(time_t *)(buf + 2) );
    char *path, *from, *subject, *digest;

    char *cp = buf + 2 * sizeof(time_t) + sizeof(size_t);
    path = cp;
    while(*cp) cp++;
    from = ++cp;
    while(*cp && cp < buf + blksize - 1) cp++;
    subject = ++cp;
    while(*cp && cp < buf + blksize - 1) cp++;
    digest = cp + 1;
    
    char datebuf[MAXDATESTR + 1];
    //    out << "* docid = " << result.docid << endl;
    out << '[' << path << ']';
    out << " (score: " << (int)result.weight;
    out << ", size: " << sz << " bytes)" << endl;
    out << "  Date: " << rfc822date(datebuf, MAXDATESTR, tm) << endl;
    out << "  From: " << from << endl;
    out << "  Subject: ";
    if(subject)
        cout << subject << endl;
    else cout << "(no subject)" << endl;
    out << ">> " << digest << ".." << endl;
    out << endl; 
}


void
MailDescription::printHtml(ostream &out, const Result &result){
    char url[blksize + 1];
    char author[blksize + 1];
    char title[blksize + 1];
    char digest[blksize + 1];
    char datebuf[MAXDATESTR + 1];
    time_t date, lastmod;
    size_t sz;

    getblock(result.docid, date, lastmod, sz, url, author, title, digest);
    char *escaped_url = escape_url(url);
    char buf[blksize + 1];

    out << "<!-- id:" << result.docid << "-->" << endl;
    out << "<DT><B><A HREF=\"" << escaped_url << "\">";
    if(title[0]){
        out << html_entitize(buf, title);
    }else out << "(no subject)";
    out << "</A></B>\n";
    out << " <SMALL>(score: " << (int)result.weight << ")</SMALL>\n";
    out << "<DD>";
    if(author[0])
        out << "  From: <TT>" << html_entitize(buf, author) << "</TT><BR>\n";
    out << "  Date: " << rfc822date(datebuf, MAXDATESTR, lastmod) << "<BR>";
    out << "&gt;&gt; " << html_entitize(buf, digest) << "..<BR>\n";
    out << "<SMALL><NOBR><A HREF=\"" << escaped_url << "\">"
        << escaped_url << "</A></NOBR> \n";
    out << "<NOBR>( " << sz << " bytes)</NOBR></SMALL><BR><BR>";
    out << endl; 
    delete []escaped_url;
}
