// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Indexer.cc,v 0.12 1998-02-09 22:44:47+09 harada Exp $
//

#include <fstream.h>
extern "C" {
#include <stdio.h>              // for rename(2) ...
#include <assert.h>
#include <limits.h>
#include <unistd.h>
}
#include "LexiconFile.h"
#include "CachedIndexFile.h"
#include "Indexer.h"
#include "typedef.h"
#include "Word.h"
#include "HeapBag.h"
#include "XMap.h"
#include "kcnvlib.h"
#include "config.h"
#include "settemplate.h"

Indexer::Indexer(WordBreaker &wb, const Unifier &unif)
    : breaker(wb), unifier(unif)
{
}

Indexer::~Indexer(){
}

void
Indexer::printDebug(ostream &out) const{
    out << "-- Indexer --" << endl;
    out << "(no useful info)" << endl;
}

int
Indexer::buildindex(LexiconFile &lexfile, IndexFile &indexfile,
                    HashBag<valueSet<pos_t> > &hashbuf){
    Word *wp = NULL;
    valueSet<pos_t> *vp = NULL;
    HeapBag<Word> heap(hashbuf.size());
    Word *sorted[hashbuf.size()];

    int keyNum = 0;
    hashbuf.rewind();
    while(hashbuf.next((HashKey *&)wp, vp)){
        keyNum++;
        // cout << *wp << " => " << *vp << endl;
        assert(wp->length() != 0);
        heap.insert(wp);
    }
    assert(keyNum == hashbuf.size());

    heap.sort(sorted);
    lexfile.prepareAppend();
    for(int i = 0; i < hashbuf.size(); i++){
        LexiconEntry entry(*sorted[i]);
        valueSet<pos_t> &value = hashbuf[*sorted[i]];
        // cout << *sorted[i] << " => " << value << endl;
        assert(value.size() > 0);
        entry.store(indexfile, value, value.size());
        lexfile.appendEntry(entry);
    }
    
    return keyNum;
}

// typeoffset[1]typeoffset[XMap::TYPENUM]ޤǤϤ롣
// typeoffset[0]ϽϤʤ
void
Indexer::output_mapentry(ostream &xmapfile, pos_t base, size_t typeoffset[]){
    short_size_t typerange[XMap::TYPENUM];
    size_t lastoffset = 0;
    int i;
    for(i = 0; i < XMap::TYPENUM; i++){
        if(typeoffset[i + 1] == 0)
            typerange[i] = 0;
        else{
            typerange[i] = typeoffset[i + 1] - lastoffset;
            lastoffset = typeoffset[i + 1];
        }
    }
    xmapfile.write(&base, sizeof(base));
    xmapfile.write(typerange, sizeof(typerange));
    // Ƥ
    for(i = 0; i < XMap::TYPENUM + 1; i++)
        typeoffset[i] = 0;
}

void
Indexer::output_last_mapentry(ostream &xmapfile, pos_t offset){
    static short int scores[XMap::TYPENUM]
        = { WEIGHT_URL, WEIGHT_TITLE, WEIGHT_AUTHOR,
            WEIGHT_KEYWORD, WEIGHT_TEXTW1, WEIGHT_TEXTW2 };
    xmapfile.write(&offset, sizeof(offset));
    xmapfile.write(scores, sizeof(scores));
}

void
Indexer::deentitize(char *dst, const char *src){
    const char *cp = src;
    while(*cp){
        if(*cp == '&'){
            if(0 == strncmp(cp, "&gt;", 4)){
                *dst++ = '>';
                cp += 4;
            }else if(0 == strncmp(cp, "&lt;", 4)){
                *dst++ = '<';
                cp += 4;
            }else if(0 == strncmp(cp, "&amp;", 5)){
                *dst++ = '&';
                cp += 5;
            }else if(0 == strncmp(cp, "&nbsp;", 6)){
                *dst++ = ' ';
                cp += 6;
            }else if(0 == strncmp(cp, "&quot;", 6)){
                *dst++ = '"';
                cp += 6;
            }else if(0 == strncmp(cp, "&apos;", 6)){
                *dst++ = '\'';
                cp += 6;
            }else{
                *dst++ = *cp++;
            }
        }else{
            *dst++ = *cp++;
        }
    }
    *dst = '\0';
}

int
Indexer::processText(istream &source, LexiconFile &lexfile,
                     IndexFile &indexfile, fstream &xmapfile, pos_t &offset){
    // Ⱦѥľ'-'¿ʤɡˤƤޤΤۤ
    // Ĺʤ뤳Ȥ⤢
    char linebuf[MAXSOURCELINELENGTH + 1024];
    char unifiedline[MAXSOURCELINELENGTH + 512];
    size_t typeoffset[XMap::TYPENUM + 1];
    int type = TYPE_TEXTW3;
    int docid = -1;
    pos_t start_offset = offset;

    HashBag<valueSet<pos_t> > words(MAXBUFFERINGWORDS);
    linebuf[MAXSOURCELINELENGTH - 1] = '\0';

    for(int lineno = 0 ; ; lineno++){
        source.getline(linebuf, MAXSOURCELINELENGTH, '\n');
        if(source.fail()){
            if(source.eof()) break;
            // linebufǥޥХʸǤ줿ν
            // cerr << "fail! line = \"" << linebuf << "\"" << endl;
            source.clear();
            // source.clear(ios::failbit); Τۤ?
            char *cp = linebuf;
            while(*cp){
                if(Iskanji1(*cp) || Ishkana1(*cp)){
                    if(*(cp + 1) == '\0'){
                        source.putback(*cp);
                        *cp = '\0';
                        break;
                    }
                    if(Iskanji2(*(cp + 1))) cp +=  2;
                    else if(Ishkana2(*(cp + 1))) cp += 2;
                    else {
                        PRINTERRMSG("Abnormal Kanji sequence found. ignore.");
                        *cp++ = ' ';
                        *cp++ = ' ';
                    }
                } else
                    cp++;
            }
        }
        // cout << "lineno = " << lineno << " line=\"" << linebuf;
        // cout << "\" size: " << words.size() << endl;
        if(linebuf[0] == '\0') continue;
        if(linebuf[0] == '<'){
            if(0 == strncasecmp(linebuf, "<DOC HREF=\"", 11)){
                start_offset = offset;
                //cout << "start_offset = " << start_offset << endl;
                char *url = linebuf + 11;
                char *q = strchr(url, '"');
                if(q) *q = '\0';
                deentitize(unifiedline, url);
                (void)breaker.breakText(words, unifiedline,
                                        offset, WordBreaker::FORINDEX);
                cout << '[' << unifiedline << flush;
                for(type = 0; type < XMap::TYPENUM + 1; type++)
                    typeoffset[type] = 0;
                typeoffset[TYPE_URL] = offset - start_offset;
                docid++;
                type = TYPE_TEXTW3;
            }else if(0 == strncasecmp(linebuf, "<TEXT WEIGHT=\"", 14)){
                int weight = atoi(linebuf + 14);
                if(weight <= 0 || 3 < weight) weight = 3;
                switch (weight){
                case 3: type = TYPE_TEXTW3; break;
                case 2: type = TYPE_TEXTW2; break;
                case 1: type = TYPE_TEXTW1; break;
                default: break;
                }
            }else if(0 == strncasecmp(linebuf, "<KEYWORD", 8)){
                type = TYPE_KEYWORD;
            }else if(0 == strncasecmp(linebuf, "<TITLE", 6)){
                type = TYPE_TITLE;
            }else if(0 == strncasecmp(linebuf, "<AUTHOR", 7)){
                type = TYPE_AUTHOR;
            }else if(0 == strncmp(linebuf, "</", 2)){
                typeoffset[type] = offset - start_offset;
                if(0 == strncasecmp(linebuf, "</DOC", 5)){
                    output_mapentry(xmapfile, start_offset, typeoffset);
                    cout << ']'<< endl;
                }
                type = TYPE_TEXTW3;
                offset++;
            }
            continue;
        }
        unifier.unify(unifiedline, linebuf);
        deentitize(linebuf, unifiedline);

        // ʬ
        (void)breaker.breakText(words, linebuf,
                                offset, WordBreaker::FORINDEX);
        if(words.size() > MAXBUFFERINGWORDS){
            buildindex(lexfile, indexfile, words);
            return 1;           // ޤ³
        }
    }
    if(words.size() > 0){
        buildindex(lexfile, indexfile, words);
    }
    offset++;
    output_last_mapentry(xmapfile, offset);
    return 0;                   // Ϥ
}


void
Indexer::makefilename(char *lexfilename, char *idxfilename,
                      const char *basename, int nth){
    const char *cp = basename;
    char *dp = lexfilename;
    while((*dp = *cp++)) dp++;
    *dp++ = '.';
    *dp++ = '0' + nth / 100;
    *dp++ = '0' + (nth % 100) / 10;
    *dp++ = '0' + nth % 10;
    *dp = '\0';
    strcpy(idxfilename, lexfilename);
    strcat(idxfilename, ".idx");
    strcat(lexfilename, ".lex");
}

int
Indexer::makeIndices(istream &source, const char *basename,
                     fstream &xmapfile, pos_t &offset){
    char lexfilename[PATH_MAX];
    char idxfilename[PATH_MAX];
    int no;

    xmapfile.seekp(0);
    for(no = 0; no <= 999; no++){
        makefilename(lexfilename, idxfilename, basename, no);
        fstream lexfile(lexfilename, ios::in|ios::out|ios::trunc, 0666);
        fstream idxfile(idxfilename, ios::in|ios::out|ios::trunc, 0666);
        if(lexfile.good() && idxfile.good()){
            LexiconFile lexicon(lexfile, 0, 1); // !readonly, newfile
            CachedIndexFile indexfile(8, idxfile, 0, 1); // !readonly, newfile
            if(0 == processText(source, lexicon,
                                indexfile, xmapfile, offset)){
                no++;
                break;
            }
        }else{
            PRINTERRMSG("Internal error: Cannot open files");
            cerr << "  lexfilename = \"" << lexfilename << "\"\n";
            cerr << "  idxfilename = \"" << idxfilename << "\"\n";
            return 0;
        }
    }
    return no;                  // ե(Ȥ)ο
}

int
Indexer::mergeIndices(const char *newbase, const char *tmpbase, int tmpNum){
    assert(tmpNum > 0);
    if(tmpNum == 1){            // Ĥξ̾ѹ
        char newlexfilename[PATH_MAX];
        char newidxfilename[PATH_MAX];
        char tmplexfilename[PATH_MAX];
        char tmpidxfilename[PATH_MAX];

        makefilename(tmplexfilename, tmpidxfilename, tmpbase, 0);
        strcpy(newlexfilename, newbase); strcat(newlexfilename, ".lex");
        strcpy(newidxfilename, newbase); strcat(newidxfilename, ".idx");

        if(-1 == rename(tmplexfilename, newlexfilename)) PRINTERR();
        if(-1 == rename(tmpidxfilename, newidxfilename)) PRINTERR();
        
        return 0;
    }

    // tmpNumȤΥƥݥեreadonlyǰƤ˳
    int i;
    char idxfilenames[tmpNum][PATH_MAX];
    char lexfilenames[tmpNum][PATH_MAX];
    fstream *lexfiles[tmpNum];
    fstream *idxfiles[tmpNum];
    LexiconFile *lexicons[tmpNum];
    IndexFile *indices[tmpNum];

    for(i = 0; i < tmpNum; i++){
        makefilename(lexfilenames[i], idxfilenames[i], tmpbase, i);
        //cout << "lexfilenames[i] = " << lexfilenames[i] << endl;
        lexfiles[i] = new fstream(lexfilenames[i], ios::in);
        idxfiles[i] = new fstream(idxfilenames[i], ios::in);
        lexicons[i] = new LexiconFile(*lexfiles[i], 1, 0); // rdonly, newfile
        indices[i] = new CachedIndexFile(8, *idxfiles[i], 1, 0);
    }

    // ե
    char newlexpath[PATH_MAX];
    char newidxpath[PATH_MAX];
    strcpy(newlexpath, newbase); strcat(newlexpath, ".lex");
    strcpy(newidxpath, newbase); strcat(newidxpath, ".idx");
    fstream newlexfile(newlexpath, ios::in|ios::out|ios::trunc);
    fstream newidxfile(newidxpath, ios::in|ios::out|ios::trunc);
    LexiconFile newlexicon(newlexfile, 0, 1); // rdonly, newfile
    CachedIndexFile newindex(8, newidxfile, 0, 1);     // rdonly, newfile

    // ޡ
    mergeFiles(&newlexicon, &newindex, tmpNum, lexicons, indices);

    for(i = 0; i < tmpNum; i++){
        delete indices[i];
        delete idxfiles[i];
        delete lexicons[i];
        delete lexfiles[i];
        unlink(lexfilenames[i]);
        unlink(idxfilenames[i]);
    }

    return 0;
}


int
Indexer::mergeFiles(LexiconFile *newlexicon, IndexFile *newindex, int n,
                    LexiconFile *lexicons[], IndexFile *indices[],
                    size_t *shifttbl)
{
    int i, j;
    int nwords = 0;
    
    istream *lexstreamp[n];
    LexiconEntry entries[n];

#if 0
    for(i = 0; i < n; i++){
        int listLen;
        pos_t *list;
        lexstreamp[i] = &(lexicons[i]->rewind());
        for(;;){
            *lexstreamp[i] >> entries[i];
            if(lexstreamp[i]->eof()){
                cout << "[eof]!\n";
                break;
            }
            cout << "-i = " << i << ", key = " << entries[i].getkey() << endl;
            entries[i].printDebug();
            list = entries[i].fetch(*indices[i], listLen);
            cout << "listLen = " << listLen << endl;
            assert(list);
            delete []list;
            cout << endl;
        }
    }
#endif

    for(i = 0; i < n; i++){
        lexstreamp[i] = &(lexicons[i]->rewind());
    }
    newlexicon->prepareAppend();

    int minNum;
    int minlist[n];
    int actives[n];
    int activeNum = n;

    minNum = n;
    for(i = 0; i < n; i++){
        minlist[i] = i;
        actives[i] = i;
    }

    for(;;){
        // set entries[x] to the top of corresponding lexstreamp[x];
        for(i = 0; i < minNum; i++){
            int m = minlist[i];
            *lexstreamp[m] >> entries[m];
            if(lexstreamp[m]->eof()){
                activeNum = 0;
                for(j = 0; j < n; j++)
                    if(! lexstreamp[j]->eof())
                        actives[activeNum++] = j;
            }
        }
        if(activeNum <= 1) break;
        assert(activeNum <= n);

        // find the smallest entries
        minNum = 0;
        minlist[0] = actives[0];
        for(i = 0; i < activeNum; i++){
            int a = actives[i];
            int cmp = strcmp(entries[a].getkey(),
                             entries[minlist[0]].getkey());
            // cout << "cmp " << entries[i].getkey() << " : " <<
            //    entries[minlist[0]].getkey() << endl;
            if(cmp < 0){
                minlist[0] = a;
                minNum = 1;
            }else if(cmp == 0){
                minlist[minNum++] = a;
            }
        }

        assert(minNum > 0);
        assert(minNum <= activeNum);
        // cout << "minNum = " << minNum << endl;
        // cout << "min =\"" << entries[minlist[0]].getkey() << "\"" << endl;

        // fetch data from entries specified by minlist[], 
        pos_t *lists[minNum];
        int listLens[minNum];
        int totalLen = 0;
        for(i = 0; i < minNum; i++){
            lists[i] = entries[minlist[i]].fetch(*indices[minlist[i]], listLens[i]);
#if 0
            if(lists[i] == NULL || listLens[i] <= 0
               || 0 == strcmp(entries[minlist[i]].getkey(), "")){
                cout << "listLens=" << listLens[i] << endl;
                cout << "i = " << i << " entries[minlist[i]]=\"" <<
                    entries[minlist[i]].getkey() << "\" ";
                cout << minlist[i] << endl;
                entries[minlist[i]].printDebug();
            }
#endif
            assert(lists[i] != NULL);
            assert(listLens[i] > 0);
            totalLen += listLens[i];
        }
        // and merge them
        pos_t merged[totalLen];
        int mergedLen = 0;
        if(shifttbl == NULL){   // we're merging temporal files
            pos_t tmpmerged[totalLen];
            for(i = 0; i < minNum; i++){
                mergedLen = mergeList(tmpmerged, merged, mergedLen,
                                      lists[i], listLens[i]);
                memcpy(merged, tmpmerged, sizeof(pos_t) * mergedLen);
                delete []lists[i];
            }
        }else{                  // we're merging existing files
            for(i = 0; i < minNum; i++){
                for(j = 0; j < listLens[i]; j++){
                    lists[i][j] += shifttbl[minlist[i]];
                }
                memcpy(merged + mergedLen, lists[i],
                       sizeof(pos_t) * listLens[i]);
                mergedLen += listLens[i];
                delete []lists[i];
            }
        }
        // and store as new entry
        LexiconEntry newentry(entries[minlist[0]].getkey());
        newentry.store(*newindex, merged, mergedLen);
        newlexicon->appendEntry(newentry);
        nwords++;
    }
    
    if(activeNum == 0)
        return nwords;

    assert(activeNum == 1);
    int thelast = -1;
    for(j = 0; j < n; j++)
        if(! lexstreamp[j]->eof()){
            thelast = j;
            break;
        }
    assert(thelast != -1);
    LexiconEntry entry = entries[thelast]; // 1Ĥޡ줺˻ĤäƤ
    do{
        LexiconEntry newentry(entry.getkey()); // key
        int dataLen = -1;
        pos_t *data = entry.fetch(*indices[thelast], dataLen);
        assert(data != NULL);
        assert(dataLen > 0);
        newentry.store(*newindex, data, dataLen);
        newlexicon->appendEntry(newentry);
        nwords++;
        delete []data;
        *lexstreamp[thelast] >> entry;
    }while(! lexstreamp[thelast]->eof());

    return nwords;
}


#ifdef INDEXER_DEBUG
int
main(){
    fstream morphdicfile("test.dic", ios::in);
    Patricia morphdic(morphdicfile);
    morphdicfile.close();
    if(! morphdic){
        cerr << "abort.\n";
        return 1;
    }
    fstream xmapfile("test.map", ios::in|ios::out|ios::trunc, 0666);
    WordBreaker wbreaker(morphdic);
    const Unifier unif;
    
    cout << "^^^^^\n";

    Indexer idxer(wbreaker, unif);
    pos_t offset = 1;
    int n = idxer.makeIndices(cin, "testbase", xmapfile, offset);
    cout << "indexed... n = " << n << endl;
    idxer.mergeIndices("newbase", "testbase", n);
    idxer.printDebug();
    xmapfile.close();
    cout << "^^^^^\n";

    return 0;
}
#endif
