// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Indexer.h,v 0.10 1998-02-09 22:45:05+09 harada Exp $
//
#ifndef INDEXER_H_INCLUDED
#define INDEXER_H_INCLUDED

#include <iostream.h>
#include "typedef.h"
#include "WordBreaker.h"
#include "HashBag.h"
#include "Unifier.h"
#include "IndexFile.h"
#include "LexiconFile.h"

class Indexer {
private:
    enum element_type { TYPE_TEXTW3 = 0, TYPE_URL = 1, TYPE_TITLE = 2,
                        TYPE_AUTHOR = 3, TYPE_KEYWORD = 4,
                        TYPE_TEXTW1 = 5, TYPE_TEXTW2 = 6
    };
    enum weight_by_type { WEIGHT_TEXTW3 = 8, WEIGHT_URL = 16, WEIGHT_TITLE = 9,
                          WEIGHT_AUTHOR = 9, WEIGHT_KEYWORD = 12,
                          WEIGHT_TEXTW1 = 6, WEIGHT_TEXTW2 = 4
    };
    WordBreaker &breaker;
    const Unifier &unifier;

    int buildindex(LexiconFile &lexfile, IndexFile &idxfile,
                   HashBag<valueSet<pos_t> > &hashbuf);
    static void makefilename(char *lexfilename, char *idxfilename,
                             const char *basename, int nth);
    void output_mapentry(ostream &xmapfile, pos_t base, size_t typeoffset[]);
    void deentitize(char *dst, const char *src);
    virtual void output_last_mapentry(ostream &xmapfile, pos_t offset);

public:
    Indexer(WordBreaker &wb, const Unifier &unif);
    virtual ~Indexer();
    int processText(istream &source, LexiconFile &lexfile, IndexFile &idxfile,
                    fstream &xmapfile, pos_t &offset);

    int makeIndices(istream &source, const char *basename,
                   fstream &xmapfile, pos_t &offset);
    static int mergeIndices(const char *newbase,
                            const char *tmpbase, int tmpNum);
    static int mergeFiles(LexiconFile *newlexicon, IndexFile *newindex,
                          int n, LexiconFile *lexicons[],
                          IndexFile *indices[], size_t *shifttbl = NULL);

    virtual void printDebug(ostream &out = cout) const;
};

#endif
