// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Retriever.cc,v 0.9 1997-12-27 07:23:31+09 harada Exp $
//
#include <iostream.h>
#include <fstream.h>
extern "C" {
#include <assert.h>
#include <math.h>
}
#include "config.h"
#include "kcnvlib.h"
#include "settemplate.h"
#include "LexiconEntry.h"
#include "Retriever.h"

Retriever::Retriever(LexiconFile &lex, IndexFile &idxfile, const XMap &map)
    : lexiconfile(lex), indexfile(idxfile), xmap(map)
{
}

Retriever::~Retriever(){
}


pos_t *
Retriever::searchJapaneseWord(const char *key, int &resultLen,
                              int isprefixsearch){
    pos_t *list = NULL;
    int listLen = 0;
    int p = 0;
    resultLen = 0;
    while(key[p] != '\0'){
        LexiconEntry entry;
        // cout << "key = " << key + p << endl;
        int matchLen = lexiconfile.findPrefix(key + p, entry);
        // cout << "matchLen = " << matchLen << endl;
        if(! isprefixsearch){
            if(0 == matchLen){
                if(list) delete []list;
                resultLen = 0;
                return NULL; // not found
            }
        }else{
            if(0 == matchLen || *(key + p + matchLen) == '\0'){
                int tmpLen = 0;
                pos_t *tmplist = prefixsearchWord(key + p, tmpLen);
                if(tmplist){
                    if(list){
                        int mergedLen = (tmpLen < listLen) ? tmpLen : listLen;
                        pos_t *mergedlist = new pos_t[mergedLen];
                        mergedLen = multiplyList(mergedlist, list, listLen,
                                                 tmplist, tmpLen, p);
                        delete []tmplist;
                        delete []list;
                        if(mergedLen > 0){
                            resultLen = mergedLen;
                            return mergedlist;
                        }else{  
                            delete []mergedlist;
                            resultLen = 0;
                            return NULL;    // not found
                        }
                    }else{
                        assert(p == 0);
                        resultLen = tmpLen;
                        return tmplist;
                    }
                }
            }
        }
        assert(matchLen % 2 == 0);
        if(list == NULL){
            list = entry.fetch(indexfile, listLen);
            if(list == NULL) PRINTERRMSG("Internal error.");
            resultLen = listLen;
            assert(list != NULL);
            assert(listLen > 0);
        }else{
            int tmplistLen = 0;
            int mergedLen = 0;
            pos_t *tmplist = entry.fetch(indexfile, tmplistLen);
            pos_t *mergedlist;

            if(tmplist == NULL) PRINTERRMSG("Internal error.");
            assert(tmplist != NULL);
            assert(tmplistLen > 0);
            mergedLen = (listLen < tmplistLen) ? listLen : tmplistLen;
            mergedlist = new pos_t[mergedLen];
            mergedLen = multiplyList(mergedlist, list, listLen,
                                     tmplist, tmplistLen, p);
            delete []tmplist;
            delete []list;
            if(mergedLen > 0){
                list = mergedlist;
                resultLen = listLen = mergedLen;
            }else{
                delete []mergedlist;
                resultLen = 0;
                return NULL;    // not found
            }
        }
        if(key[p + matchLen] == '\0')
            break;
        if(matchLen <= 4){
            p += 2;
        }else{
            p += matchLen;
        }
    }
    return list;
}

pos_t *
Retriever::searchWord(const char *key, int &resultLen){
    LexiconEntry entry;
    resultLen = 0;
    if(0 == lexiconfile.search(key, entry))
        return NULL;
    return entry.fetch(indexfile, resultLen);
}


// ե졼
pos_t *
Retriever::searchPhrase(char **keys, int offsets[],
                        int nkeys, int &resultLen, int isprefixsearch){
    pos_t *list = NULL;
    for(int k = 0; k < nkeys; k++){
        pos_t *tmplist;
        int tmplistLen;

        if(k == nkeys - 1)
            tmplist = search(keys[k], tmplistLen, isprefixsearch);
        else
            tmplist = search(keys[k], tmplistLen, 0);
        if(tmplist == NULL){    // not found
            if(list) delete []list;
            return NULL;
        }
        if(list == NULL){
            list = tmplist;
            resultLen = tmplistLen;
        }else{
            int mergedLen = (resultLen < tmplistLen) ? resultLen : tmplistLen;
            pos_t *mergedlist = new pos_t[mergedLen];
            mergedLen = multiplyList(mergedlist, list, resultLen,
                                     tmplist, tmplistLen, offsets[k]);
            delete []list;
            delete []tmplist;
            if(mergedLen > 0){
                list = mergedlist;
                resultLen = mergedLen;
            }else{ // not found
                delete []mergedlist;
                return NULL;
            }
        }
    }
    return list;
}


static int
compare_pos_t(const pos_t *a, const pos_t *b){
    return *a - *b;
}

pos_t *
Retriever::prefixsearchWord(const char *key, int &resultLen){
    int i, n;
    LexiconEntry **entries;

    resultLen = 0;
    entries = lexiconfile.prefixsearch(key, n);
    if(NULL == entries) return NULL; // not found

    //cout << "prefixsearch: key=" << key << endl;
    //for(i = 0; i < n; i++){
    //    cout << "  " << entries[i]->getkey() << endl;
    //}

    pos_t *list;
    if(n == 1){
        list = entries[0]->fetch(indexfile, resultLen);
    }else{
        pos_t *lists[n];
        int listLens[n];
        int totalLen = 0;
        for(i = 0; i < n; i++){
            lists[i] = entries[i]->fetch(indexfile, listLens[i]);
            assert(lists[i] != NULL);
            assert(listLens[i] > 0);
            totalLen += listLens[i];
        }
        list = new pos_t[totalLen];
        int from = 0;
        for(i = 0; i < n; i++){
            memcpy(list + from, lists[i], sizeof(pos_t) * listLens[i]);
            from += listLens[i];
            delete []lists[i];
        }
        qsort(list, totalLen, sizeof(pos_t),
              (int (*)(const void *, const void *))compare_pos_t);
        resultLen = totalLen;
    }

    for(i = 0; i < n; i++)
        delete entries[i];
    delete []entries;
    if(resultLen == 0){
        if(list) delete []list;
        return NULL;
    }
    return list;
}

// ޤޤʤθ
//      search()searchPhrase()ߤ˸ƤӤ礦
//      NPפΤ褦ܸե졼ե졼˸뤳Ȥ뤿
pos_t *
Retriever::search(const char *text, int &resultLen, int isprefixsearch){
    char keybufs[MAXWORDSINPHRASE][MAXKEYLENGTH];
    char *keys[MAXWORDSINPHRASE];
    int offsets[MAXWORDSINPHRASE];
    const char *cp = text;
    pos_t *list = NULL;
    int k = 0;
    while(*cp){
        offsets[k] = cp - text;
        char *dp = keys[k] = keybufs[k];
        if(Is_kanji_here(cp)){
            do { *dp++ = *cp++; *dp++ = *cp++; } while(Is_kanji_here(cp));
        }else{
            do { *dp++ = *cp++; } while(*cp && !Iskanji1(*cp));
        }
        *dp = '\0';
        k++;
    }
    if(k == 1){
        if(Is_kanji_here(keys[0])){
            list = searchJapaneseWord(keys[0], resultLen, isprefixsearch);
        }else{
            if(isprefixsearch)
                list = prefixsearchWord(keys[0], resultLen);
            else
                list = searchWord(keys[0], resultLen);
        }
    }else if(k > 1){
        list = searchPhrase(keys, offsets, k, resultLen, isprefixsearch);
    }
    if(resultLen == 0){
        if(list) delete []list;
        return NULL;
    }
    return list;
}


Result *
Retriever::translateList(const pos_t *list, int listLen,
                         int &resultLen, int type){
    int i;
    assert(listLen > 0);
    Result *result = new Result[listLen];
    int r = 0;
    urlid_t lastid = (urlid_t)-1;
    for(i = 0; i < listLen; i++){
        urlid_t id;
        size_t inoffset;
        int t;
        if(0 != xmap.translate(list[i], id, inoffset, t)){
            PRINTERRMSG("translation fail!");
            cerr << "-> list[" << i << "]=" << list[i] << endl;
        }else if(type == 0 || type == t){
            if(id == lastid) r--;
            result[r].docid = id;
            result[r].docsize = xmap.getdocsize(id);
            result[r].ntimes += 1;
            result[r].weight += xmap.score(t);
            r++;
            lastid = id;
        }
    }
    if(r == 0){                 // not found
        delete []result;
        return NULL;
    }
    resultLen = r;
    // calculate the average size
    double sum = 0;
    for(i = 0; i < resultLen; i++)
        sum += result[i].docsize;
    double avg = (size_t)(sum / resultLen);
    // calculate `tf * idf'
    double idf = log(2 * xmap.getdocNum() / resultLen);
    for(i = 0; i < resultLen; i++)
        if(result[i].docsize > avg / 2)
            result[i].weight *= (avg * idf) / result[i].docsize;
        else
            result[i].weight *= idf;
    return result;
}


Result *
Retriever::retrievePhrase(char **keys, int *offsets, int nkeys,
                          int &resultLen, int type){
    int listLen;
    pos_t *list = searchPhrase(keys, offsets, nkeys, listLen);
    if(list == NULL)
        return NULL;
    else{
        Result *result = translateList(list, listLen, resultLen, type);
        delete []list;
        return result;
    }
}

Result *
Retriever::retrieve(const char *text, int &resultLen,
                    int type, int isprefixsearch){
    int listLen;
    pos_t *list = search(text, listLen, isprefixsearch);
    if(list == NULL)
        return NULL;
    else{
        Result *result = translateList(list, listLen, resultLen, type);
        delete []list;
        return result;
    }
}



#ifdef RETRIEVER_DEBUG
#include <unistd.h>
#include <fcntl.h>

int main(){
    fstream idxstream("test.idx", ios::in);
    IndexFile idxfile(idxstream, 1, 0); // readonly, !newfile
    fstream lexfs("test.lex", ios::in);
    LexiconFile lexfile(lexfs, 1, 0); // readonly, !newfile
    int fd = open("test.map", O_RDONLY, 0666);
    if(fd == -1){
        PRINTERRMSG("fail to open");
        return 1;
    }
    XMap map(fd);
        
    cout << "^^^^^\n";
    Retriever retr(lexfile, idxfile, map);

    int rfound;
#if 0
    // pos_t *result = retr.search("Хץ", rfound);
    pos_t *list = retr.search("connectivity", rfound);
    if(list){
        for(int j = 0; j < rfound; j++)
            cout << list[j] << ' ';
        cout << endl;
        delete []list;
    }
#endif
    Result *result = retr.retrieve("sendmail", rfound, 0);
    if(result){
        cout << "result: ";
        for(int r = 0; r < rfound; r++){
            cout << result[r].docid << ' ';
        }
        cout << endl;
    }

    cout << "^^^^^\n";

    return 0;
}
#endif
