// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: WordBreaker.h,v 0.10 1998-01-08 08:57:53+09 harada Exp $
//
#ifndef WORDBREAKER_H_INCLUDED
#define WORDBREAKER_H_INCLUDED

extern "C" {
#include <ctype.h>
}
#include "typedef.h"
#include "HashBag.h"
#include "valueSet.h"
#include "Word.h"
#include "patricia.h"

class WordBreaker {
private:
    Patricia &morphdic;

    int entryWord(HashBag<valueSet<pos_t> > &bag, const char *word,
                  int wordLen, pos_t offset, int method);
    int is_valid_kanji(const char *kp) const;
    inline int is_valid_head(char c) const;
    inline int is_valid_tail(char c) const;
    inline int is_valid_char(char c) const;
public:
    enum limit { MAXPREFIXES = 128, MAXSTRINGLENGTH = 4096 };
    enum method { FORSEARCH = 1, FORINDEX = 2 };
    WordBreaker(Patricia &pat); // this is NOT copy constructor
    ~WordBreaker();
    int breakJapaneseText(HashBag<valueSet<pos_t> > &bag,
                          const char *text, pos_t &offset);
    int breakText(HashBag<valueSet<pos_t> > &bag, const char *text,
                  pos_t &offset, int method = FORSEARCH);
};

inline int
WordBreaker::is_valid_head(char c) const{
    return isalnum((unsigned char)c); // isalpha()Τۤ
}

inline int
WordBreaker::is_valid_char(char c) const{
    if( c == '-' || c == '+' || c == '_')
        return (int)c;
    else
        return isalnum((unsigned char)c);
}

inline int
WordBreaker::is_valid_tail(char c) const{
    if( c == '-' || c == '+' || c == '_' ) return (int)c;
    else return isalnum((unsigned char)c);
}


#endif
