// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// address translator (mapping)
// $Id: XMap.cc,v 0.9 1997-12-27 07:27:23+09 harada Exp $
//
#include <iostream.h>
#include <iomanip.h>
extern "C" {
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/mman.h>
}
#include "XMap.h"
#include "config.h"

XMap::XMap(int filedesc)
    : fd(filedesc)
{
    struct stat st;
    fstat(fd, &st);
    mapsize = st.st_size;
    entryNum = mapsize / ENTRYSIZE - 1;
    mapaddr = mmap(NULL,        // start address (hint)
                   mapsize,     // size
                   PROT_READ,   // protection
                   MAP_SHARED,  // flags
                   fd,          // file descriptor
                   (off_t)0 );         // file offset
    if((caddr_t)(-1) == mapaddr)
        PRINTERRMSG("mmap() failed.");
    maxoffset = getbase(entryNum); // 

    scores = new int[TYPENUM + 1];
    short int *p = (short int *)((char *)mapaddr +
                                 ENTRYSIZE * entryNum + sizeof(pos_t));
    scores[0] = 1;
    for(int i = 0; i < TYPENUM; i++)
        scores[i+1] = p[i];
}

XMap::~XMap(){
    if(mapaddr && -1 == munmap(mapaddr, mapsize)){
        PRINTERRMSG("munmap(2) failed.");
    }
    if(-1 == close(fd))
        PRINTERRMSG("close(2) failed.");
    delete []scores;
}

void
XMap::printDebug(ostream &out) const {
    int i;
    out << "-- XMap --" << endl;
    out << "  entryNum = " << entryNum << endl;
    out << "  TYPENUM = " << TYPENUM << " [ ";
    for(i = 0; i < TYPENUM; i++){
        out << score(i) << ' ';
    }
    out << ']' << endl;
    for(i = 0; i < entryNum; i++){
        out << " (" << i << " : " << getbase(i) << ')';
    }
    out << endl;
}

// ʬõ. bsearch(3)Ǥ⤤
int
XMap::findEntry(pos_t pos) const{
    if(pos > maxoffset)
        return -1; // not found

    int left = 0;
    int right = entryNum;
    while(right > left + 1){
        int median = (left + right) / 2;
        if(pos < getbase(median)){
            right = median;
        }else{
            assert(getbase(median) <= pos);
            left = median;
        }
    }
    if(getbase(left) > pos)
        return -1;              // not found
    return left;
}


// pos_tΰ־Ѵ
//  0顼0ʳ
// и֤ urlid, offset, type ɽ
int
XMap::translate(pos_t pos, urlid_t &urlid, size_t &offset, int &type) const{
    int id = findEntry(pos);
    if(id == -1) return -1;
    urlid = (urlid_t)id;
    short_size_t *p;
    p = (short_size_t *)((char *)mapaddr + id * ENTRYSIZE + sizeof(pos_t));
    pos_t tmpoffset = getbase(id);
    offset = pos - tmpoffset;
    type = 0;
    for(int t = 1; t <= TYPENUM; t++, p++){
        tmpoffset += *p;
        if(pos < tmpoffset){
            type = t;
            break;
        }
    }

    return 0;
}


#ifdef XMAP_DEBUG
// g++ -O -g -DXMAP_DEBUG XMap.cc
int main(int argc, char *argv[]){
    int fd = open("test.map", O_RDONLY, 0666);
    if(fd == -1){
        PRINTERRMSG("fail to open");
        return 1;
    }
    XMap map(fd);
    if(! map){
        PRINTERRMSG("cannot open translation map");
        return 2;
    }

    urlid_t id = (urlid_t)-1;
    size_t inoffset = (size_t)-1;
    int type = -1;
    //map.translate(1494839, id, inoffset, type);
    if(map.translate(21505, id, inoffset, type)){
        cout << "translation failed!\n";
    }
    // map.translate(1, id, inoffset, type);
    cout << "id = " << id << " inoffset = " << inoffset
         << " type = " << type << endl;

    map.printDebug();

    return 0;
}

#endif
