// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: findex.cc,v 0.10 1997-12-29 14:22:42+09 harada Exp $
//
#include <iostream.h>
extern "C" {
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
}
#include "config.h"
#include "typedef.h"
#include "Indexer.h"

char *
mystrdup(const char *str){
    int l = strlen(str);
    char *s = new char[l + 1];
    strcpy(s, str);
    return s;
}

void
printusage(){
    cout << FREYABANNER << endl
         << "Usage: findex [-d dictfile] [-o indexname] source" << endl;
}

int
main(int argc, char *argv[]){
    const char *dictpath = DEFAULT_MORPHDIC_PATH;
    const char *outputbase = "default";
    extern char *optarg;
    extern int optind; // opterr, optopt;

    int ch;
    while((ch = getopt(argc, argv, "d:o:v?")) != EOF){
        switch(ch){
        case 'd':
            dictpath = mystrdup(optarg);
            break;
        case 'o':
            outputbase = mystrdup(optarg);
            break;
        case 'v':
        case '?':
        default:
            printusage();
            return 1;
        }
    }
    argc -= optind;
    argv += optind;

    if(argc != 1){
        printusage();
        return 1;
    }

    const char *filepath = argv[0];
#ifndef NDEBUG
    cout << "dictpath = " << dictpath << endl;
    cout << "filepath = " << filepath << endl;
    cout << "outputbase = " << outputbase << endl;
#endif

    // setup the word breaker
    fstream morphdicfile(dictpath, ios::in);
    Patricia morphdic(morphdicfile);
    morphdicfile.close();
    if(! morphdic){
        PRINTERRMSG("Error: invalid morphdic.");
        cerr << "  dictpath = \"" << dictpath << "\"\n";
        return 1;
    }
    WordBreaker wbrk(morphdic);

    // setup the indexer
    const Unifier unif;
    Indexer idxer(wbrk, unif);
    pos_t offset = 1;

    // prepare xmap
    char xmappath[PATH_MAX];
    strcpy(xmappath, outputbase); strcat(xmappath, ".map");
    fstream xmapfile(xmappath, ios::in|ios::out|ios::trunc, 0666);
    if(! xmapfile.good()){
        PRINTERRMSG("cannot create a translation map file.\n");
        cerr << "  xmappath = \"" << xmappath << "\"\n";
        return 2;
    }
    if(errno == ENOENT)         // clear "No such file or directory"
        errno = 0;

    // Let's index!
    char tmpbase[PATH_MAX];
    strcpy(tmpbase, outputbase); strcat(tmpbase, "_");
    int tmpfileNum;
    if(0 == strcmp(filepath, "-")){
        tmpfileNum = idxer.makeIndices(cin, tmpbase, xmapfile, offset);
    }else{
        fstream source(filepath, ios::in);
        if(! source.good()){
            PRINTERRMSG("cannot open source text. ");
            cerr << "  filepath = \"" << filepath << "\"\n";
            return 3;
        }
        tmpfileNum = idxer.makeIndices(source, tmpbase, xmapfile, offset);
        source.close();
    }
    xmapfile.close();
    idxer.mergeIndices(outputbase, tmpbase, tmpfileNum);

    return 0;
}
