// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: fmerge.cc,v 0.10 1998-01-09 22:50:42+09 harada Exp $
//
#include <iostream.h>
extern "C" {
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
}
#include "config.h"
#include "typedef.h"
#include "XMap.h"
#include "CachedIndexFile.h"
#include "Indexer.h"

char *
mystrdup(const char *str){
    int l = strlen(str);
    char *s = new char[l + 1];
    strcpy(s, str);
    return s;
}

void
printusage(){
    cout << FREYABANNER << endl
         << "Usage: fmerge newindex index [index ...]" << endl;
}


// merge .dsc files, and set the number of entries contained in each
int
mergedsc(ostream &newdscfile, char *dscfilenames[], int *ndocs, int n){
    cout << "merging descriptions " << flush;
    newdscfile.seekp(0);
    for(int i = 0; i < n; i++){
        fstream dscfile(dscfilenames[i], ios::in);
        if(! dscfile.good()){
            cout << endl;
            PRINTERRMSG("cannot open .dsc file.");
            cerr << "  filename = \"" << dscfilenames[i] << '"' << endl;
            return i + 1;
        }
        struct stat fstatus;
        fstat(dscfile.rdbuf()->fd(), &fstatus);
        off_t filesize = fstatus.st_size;
        ndocs[i] = filesize / DSCBLOCKSIZE;
        
        cout << '.' << flush;
        char buf[DSCBLOCKSIZE];
        dscfile.seekg(0);
        for(int d = 0; d < ndocs[i]; d++){
            dscfile.read(buf, DSCBLOCKSIZE);
            newdscfile.write(buf, DSCBLOCKSIZE);
        }
        dscfile.close();
    }
    cout << "ok" << endl;

    return 0;
}

// merge .map files, and set the number of entries contained in each
int
mergemap(ostream &newmapfile, char *mapfilenames[],
         int *ndocs, size_t *shifttbl, int n){
    cout << "merging translation maps " << flush;
    const int entrysize = sizeof(size_t) +
        sizeof(short_size_t) * XMap::TYPENUM;
    char entry[entrysize];
    size_t newlastoffset = 0;
    newmapfile.seekp(0);
    for(int i = 0; i < n; i++){
        fstream mapfile(mapfilenames[i], ios::in);
        if(! mapfile.good()){
            cout << endl;
            PRINTERRMSG("cannot open .map file.");
            cerr << "  filename = \"" << mapfilenames[i] << '"' << endl;
            return i + 1;
        }
        struct stat fstatus;
        fstat(mapfile.rdbuf()->fd(), &fstatus);
        off_t filesize = fstatus.st_size;
        ndocs[i] = filesize / entrysize - 1;
        
        cout << '.' << flush;
        mapfile.seekg(0);
        for(int d = 0; d < ndocs[i]; d++){
            mapfile.read(entry, entrysize);
            *(size_t *)entry += newlastoffset;
            newmapfile.write(entry, entrysize);
        }
        mapfile.read(entry, entrysize);
        shifttbl[i] = newlastoffset;
        newlastoffset += *(size_t *)entry;
    }
    *(size_t *)entry = newlastoffset;
    newmapfile.write(entry, entrysize);
    cout << "ok" << endl;

    return 0;
}

int
main(int argc, char *argv[]){
    int i;
    if(argc <= 2){
        printusage();
        return 1;
    }
    
    int srcNum = argc - 2;
    if(srcNum > 16){
        PRINTERRMSG("too many indices.");
        return srcNum;
    }

    char *lexfilenames[srcNum];
    char *idxfilenames[srcNum];
    char *mapfilenames[srcNum];
    char *dscfilenames[srcNum];
    for(i = 0; i < srcNum; i++){
        if(0 == strcmp(argv[i + 2], argv[1])){
            PRINTERRMSG("output files cannot override the input files!");
            cerr << "  output = " << argv[1]
                 << ", input = " << argv[i + 2] << endl;
            return i + 1;
        }
        lexfilenames[i] = new char[strlen(argv[i + 2]) + 5];
        idxfilenames[i] = new char[strlen(argv[i + 2]) + 5];
        mapfilenames[i] = new char[strlen(argv[i + 2]) + 5];
        dscfilenames[i] = new char[strlen(argv[i + 2]) + 5];
        strcpy(lexfilenames[i], argv[i+2]); strcat(lexfilenames[i], ".lex");
        strcpy(idxfilenames[i], argv[i+2]); strcat(idxfilenames[i], ".idx");
        strcpy(mapfilenames[i], argv[i+2]); strcat(mapfilenames[i], ".map");
        strcpy(dscfilenames[i], argv[i+2]); strcat(dscfilenames[i], ".dsc");
    }
    
    char newlexfilename[PATH_MAX];
    char newidxfilename[PATH_MAX];
    char newmapfilename[PATH_MAX];
    char newdscfilename[PATH_MAX];
    strcpy(newlexfilename, argv[1]); strcat(newlexfilename, ".lex");
    strcpy(newidxfilename, argv[1]); strcat(newidxfilename, ".idx");
    strcpy(newmapfilename, argv[1]); strcat(newmapfilename, ".map");
    strcpy(newdscfilename, argv[1]); strcat(newdscfilename, ".dsc");

    fstream newlexfs(newlexfilename, ios::in|ios::out|ios::trunc);
    fstream newidxfs(newidxfilename, ios::in|ios::out|ios::trunc);
    fstream newmapfs(newmapfilename, ios::in|ios::out|ios::trunc);
    fstream newdscfs(newdscfilename, ios::in|ios::out|ios::trunc);
    if(! (newlexfs.good() && newidxfs.good()
          && newmapfs.good() && newdscfs.good())){
        PRINTERRMSG("Something bad occured while creating output files.");
        return 1;
    }

    int ndocsdsc[srcNum];
    int ndocsmap[srcNum];       // this should be same as ndocsdsc
    size_t shifttbl[srcNum];
    if(0 != mergedsc(newdscfs, dscfilenames, ndocsdsc, srcNum))
        return 2;
    newdscfs.close();
    if(0 != mergemap(newmapfs, mapfilenames, ndocsmap, shifttbl, srcNum))
        return 2;
    newmapfs.close();
    for(i = 0; i < srcNum; i++){
        if(ndocsdsc[i] != ndocsmap[i]){
            PRINTERRMSG("inconsistency found between desc. and map.");
            cerr << "  description:\"" << dscfilenames[i] << "\", "
                 << ndocsdsc[i] << " entries." << endl;
            cerr << "  translation map:\"" << mapfilenames[i] << "\", "
                 << ndocsmap[i] << " entries." << endl;
            return i + 1;
        }
    }

    fstream *lexfiles[srcNum];
    fstream *idxfiles[srcNum];
    LexiconFile *lexicons[srcNum];
    IndexFile *indices[srcNum];
    for(i = 0; i < srcNum; i++){
        lexfiles[i] = new fstream(lexfilenames[i], ios::in);
        if(! lexfiles[i]->good()){
            PRINTERRMSG("cannot open lexicon file.");
            cerr << "  path = \"" << lexfilenames[i] << '"' << endl;
            return i + 1;
        }
        idxfiles[i] = new fstream(idxfilenames[i], ios::in);
        if(! idxfiles[i]->good()){
            PRINTERRMSG("cannot open index file.");
            cerr << "  path = \"" << idxfilenames[i] << '"' << endl;
            return i + 1;
        }
        lexicons[i] = new LexiconFile(*lexfiles[i], 1, 0);
        indices[i] = new CachedIndexFile(8, *idxfiles[i], 1, 0);

    }

    LexiconFile newlexicon(newlexfs, 0, 1);
    CachedIndexFile newindex(8, newidxfs, 0, 1);
    cout << "merging lexicons and indexfiles ..." << flush;
    Indexer::mergeFiles(&newlexicon, &newindex, srcNum,
                        lexicons, indices, shifttbl);
    cout << "ok" << endl;

    for(i = 0; i < srcNum; i++){
        delete lexicons[i];
        delete indices[i];
        delete lexfiles[i];
        delete idxfiles[i];
        delete lexfilenames[i];
        delete idxfilenames[i];
        delete mapfilenames[i];
        delete dscfilenames[i];
    }

    return 0;
}
